#!/bin/python3
#%%
import os
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier
import json

from tools.data_processing.basic_data_tools import get_hash, pickle_dump, pickle_load, silence_pickle_dump, silence_pickle_load
from sklearn.linear_model import LinearRegression
from xgboost import XGBClassifier

from tools.std_lib.log import ColorfulLogger
from tools.data_processing.basic_data_tools import dict_to_plain
from tools.data_processing.basic_data_tools import get_hash
import json
from tools.std_lib.basic import get_timestamp
from tools.modeling.training_tools_v2 import easy_make_classification, find_dir_with_prefix

from tools.modeling.training_tools_v2 import BinaryClassifierTrainer


X_df, y_df = easy_make_classification(
    n_sample=100000,
    weights=[.6, .4],
    n_features=20,
    )
# X_df.iloc[::2, :] = float('nan')

log = ColorfulLogger(basename='my_train')

# train_idx = y_df['data_split'] == 'train'

# train_X_df = X_df.loc[train_idx, :]
# train_y_s = y_df.loc[train_idx, 'y'].astype(int)


# trainer = BinaryClassifierTrainer(
#     X_train_df=train_X_df,
#     y_train_s=train_y_s,
#     classifier_name='XGB',
#     classifier_params={
#         'verbosity': 2,
#         'n_estimators': 100,
#         'max_depth': 4,
#         'learning_rate': .01,
#         # 'random_state': 0,
#         },
#     prefix = 'p9527-S13-dim114514',
#     cache_dir='./cache',
#     )

# trainer.run()



#%%


class BinaryClassifier(object):
    '''
    prefix: (defined by user)
    uuid_hex: (auto)
    exp_name: prefix-XGB_1234567890abcdef
    save_name: prefix-XGB_1234567890abcdef-20250413Sun0101
    predict_proba_name: (=prefix, default)

    exp_name: prefix-X[S01_maas_r1+S02+S03+S04_bp]-XGB_12345678
    save_name: prefix-X[S01_maas_r1+S02+S03+S04_bp]-XGB_12345678-20250413Sun0101

    '''
    def __init__(self,
                 X_df: pd.DataFrame,
                 y_df: pd.DataFrame,
                 features : list = None,
                 fillna: object = None,
                 classifier_name = 'XGB',
                 classifier_params = {},
                 prefix = None,
                 cache_dir = None,
                 tag = None,
                 ):
        self.X_df = X_df
        self.y_df = y_df
        self.features = features
        self.fillna = fillna
        self.classifier_name = classifier_name
        self.classifier_params = classifier_params
        self.prefix = prefix
        self.cache_dir = cache_dir
        self.tag = tag
        self.clf = None
        return None
    

    def X_df_fillna(self):
        if self.classifier_name in ['LR']:
            if self.fillna is None:
                train_idx = y_df['data_split'] == 'train'
                self.fillna = self.X_df.loc[train_idx, :].mean().to_dict()
            else:
                raise ValueError('fillna must be None when classifier_name = LR.')
            self.X_df = self.X_df.fillna(self.fillna)
            if self.X_df.isna().any().any():
                raise ValueError('fillna failed when LR.')
        else:
            if self.fillna is None:
                self.fillna = {k: float('nan') for k in self.X_df.columns}
            elif isinstance(self.fillna, str):
                if self.fillna == 'train_mean':
                    train_idx = y_df['data_split'] == 'train'
                    self.fillna = self.X_df.loc[train_idx, :].mean().to_dict()
                    self.X_df = self.X_df.fillna(self.fillna)
                    if self.X_df.isna().any().any():
                        raise ValueError('fillna failed when not LR.')
                else:
                    raise ValueError(f'self.fillna = {self.fillna} is not supported.')
            else:
                raise ValueError(f'self.fillna = {self.fillna} is not supported.')
        return None
    

    def get_clf_hash(self):
        train_info = {
            'X': {
                'index': self.X_df.index.tolist(),
                'columns': self.X_df.columns.tolist()
            },
            'y': self.y_df.to_dict(),
            'classifier_name': self.classifier_name,
            'classifier_params': self.classifier_params,
            'fillna': self.fillna,
        }
        train_hash = get_hash(json.dumps(train_info, sort_keys=True))
        return train_hash
    
    def get_exp_name(self):
        train_hash = self.get_clf_hash()
        exp_name = f'{self.classifier_name}_{train_hash[:16]}'
        if self.prefix is not None:
            exp_name = f'{self.prefix}-{exp_name}'
        log.info(f'exp_name = \'{exp_name}\'', func = 'clf.get_exp_name()')
        return exp_name


    def check_and_fix(self):
        if not self.X_df.index.equals(self.y_df.index):
            raise ValueError("X_df and y_df must have the same index")
        if y_df['y'].nunique() < 2:
            raise ValueError("y_df must have at least two unique values")
        if y_df['y'].isna().any():
            raise ValueError("y_df must not contain any NaN values")
        
        if self.features is None:
            self.features = self.X_df.columns.to_list()
        else:
            self.X_df = self.X_df[self.features]

        self.X_df_fillna()
        self.exp_name = self.get_exp_name()
        self.Trainer = BinaryClassifierTrainer
        # self.evaluator = BinaryEvaluator

        return None
    
    
    def set_trainer(self, trainer: object):
        self.trainer = trainer
        return None
    
    def set_evaluator(self, evaluator: object):
        self.evaluate = evaluator
        pass
 
    def exists(self):
        flag = False
        cache_dir = self.cache_dir
        if cache_dir is not None and isinstance(cache_dir, str):
            if self.prefix is None:
                prefix_path = cache_dir
            else:
                prefix_path = os.path.join(cache_dir, self.prefix)
            if os.path.exists(prefix_path) and os.path.isdir(prefix_path):
                exp_name = self.exp_name
                save_path = find_dir_with_prefix(exp_name, prefix_path)
                if save_path is not None:
                    self.save_path = save_path
                    flag = True
        log.info(f'flag = {flag}', func = 'clf.exists()')
        return flag
    
    def train(self):
        X_df = self.X_df
        y_df = self.y_df
        train_idx = y_df['data_split'] == 'train'
        X_train_df = X_df.loc[train_idx, :]
        y_train_s = y_df.loc[train_idx, 'y']
        trainer = self.Trainer(
            X_train_df = X_train_df,
            y_train_s = y_train_s,
            classifier_name = self.classifier_name,
            classifier_params = self.classifier_params,
            prefix  = self.prefix,
            cache_dir = self.cache_dir,
        )
        trainer.run()
        self.clf = trainer.clf
        return None

    def predict(self):
        X_np = X_df.to_numpy()
        proba_df = self.y_df.copy()
        proba_df['proba'] = self.clf.predict_proba(X_np)[:, 1]
        self.proba_df = proba_df
        return None

    def evaluate(self):
        pass

    def dump(self):
        pass

    def load(self):
        pass

    def run(self):
        self.check_and_fix()
        if not self.exist():
            self.fillna()
            self.train()
            self.predict()
            self.evaluate()
            self.dump()
        self.load()
        return None

    pass # end

binary_classifier = BinaryClassifier(
    X_df = X_df,
    y_df = y_df,
    classifier_name = 'XGB',
    fillna = 'train_mean',
    classifier_params = {
        # 'max_iter': 100, 
        },
    prefix = 'prefix',
    cache_dir = './cache'
)
# binary_classifier.X_df_fillna()
# binary_classifier.get_clf_hash()
binary_classifier.check_and_fix()
binary_classifier.exp_name

binary_classifier.train()

binary_classifier.predict()



#%%

proba_df = binary_classifier.proba_df


#%%
from tools.modeling.evaluation_tools import series_auc_and_ks_score

series_auc_and_ks_score(proba_df['y'], proba_df['proba'])




















#%%


# class BinaryClassifierGridSearch(object):
#     '''
#     '''
#     def __init__(self,
#                  X_df: pd.DataFrame,
#                  y_df: pd.DataFrame,
#                  fillna: object = float('nan'),
#                  classifier = 'XGBClassifier',
#                  classifier_params_list = {},
#                  prefix = None,
#                  cache_dir = None,
#                  tag_json = None,
#                  ):
#         return None


#%%

import json


def flat_to_nested(flat_dict: dict):
    s = flat_dict.copy()
    nested_dict = {}
    for key, value in s.items():
        parts = key.split('.')
        current = nested_dict
        for part in parts[:-1]:
            if part not in current:
                current[part] = {}
            current = current[part]
        current[parts[-1]] = value
    return nested_dict

params = {
    'main.prefix': 'p17-S03-dim1700',
    'main.results_path': './path/to/resutls/',
    
    'data_split.random_state': 0,
    'feature.to_str_func': 'sha256_8',
    'feature.selected': None,
    'classifier.name': 'XGBClassifier',
    'classifier.kwargs.max_depth': 4,
    'classifier.kwargs.learning_rate': 0.1,
    'classifier.kwargs.random_state': 0,
}

hyperopt_params = {
    'main.prefix': 'p17-S03-dim1700',
    'main.results_path': './path/to/resutls/',
    
    'data_split.random_state': None,
    'feature.selected': [],

    'hyperopt.k_fold': None,
    'hyperopt.space': None,
    'hyperopt.evaluator': None,
    'hyperopt.alpha': None,
    
    # 'classifier.name': 'XGBClassifier',
    # 'classifier.kwargs.max_depth': 4,
    # 'classifier.kwargs.learning_rate': 0.1,
    # 'classifier.kwargs.random_state': 0,
}

print(json.dumps(flat_to_nested(params), indent=4, ensure_ascii=False))




#%%

'''
results_path/prefix/train/
results_path/prefix/clf/
results_path/prefix/hyperopt/

'''


# class BinaryPredictor(object):
#     '''
#     '''
#     def __init__(self, 
#                  X_df: pd.DataFrame,
#                  y_df: pd.DataFrame,
#                  prefix = None,
#                  cache_dir = None,
#                  ):
#         return None


# class BinaryEvaluator(object):
#     '''
#     '''
#     def __init__(self,
#                  X_df: pd.DataFrame,
#                  y_df: pd.DataFrame,
#                  fillna: object = float('nan'),
#                  classifier = 'XGBClassifier',
#                  ):
#         return None

# def fix_classifier(classifier: str):

    # if classifier.lower() in ['xgbclassifier','xgb']:
    #     clf = XGBClassifier
    # elif classifier.lower() in ['logisticregression','lr','logistic']:
    #     clf = LogisticRegression
    # # elif classifier.lower() in ['lgbm','lgbmclassifier']:
    # #     clf = lgbmclassifier
    # # elif classifier.lower() in ['gb','gradientboostingclassifier','gradientboosting']:
    # #     clf = 'GradientBoostingClassifier'
    # # elif classifier.lower() in ['hgb','histgradientboostingclassifier','histgradientboosting']:
    # #     clf = 'HistGradientBoostingClassifier'
    # else:
    #     t_print('No clf found.')

    # return clf





# class BinaryClassifierEvaluator(object):

#     def __init__(self, 
#                  y_df: pd.DataFrame,
#                  predict_proba: pd.Series
#                  ):
#         pass

#     def evaluate(self):
#         pass


# class BinaryClassifier(object): # Once

#     def __init__(self,
#                  X_df: pd.DataFrame,
#                  y_df: pd.DataFrame,
#                  main_params: dict = {},
#                  data_split_params: dict = {},
#                  feature_params: dict = {},
#                  hyperopt_params: dict = None,
#                  classifier_params: dict = None,
#                  ):
#         self.X_df = X_df
#         self.y_df = y_df
#         self.main_params = main_params
#         self.data_split_params = data_split_params
#         self.feature_params = feature_params
#         self.hyperopt_params = hyperopt_params
#         self.classifier_params = classifier_params
#         return None
    
#     def fix_feature_params(self):

#         if 'to_str_func' not in self.feature_params:
#             self.feature_params['to_str_func'] = 'sha256_8'
#         if 'selected' not in self.feature_params:
#             self.feature_params['selected'] = None

#         if self.feature_params['selected'] is None:
#             self.feature_params['selected'] = sorted(self.X_df.columns.to_list())
#         else:
#             self.feature_params['selected'] = sorted(self.feature_params['selected'])

#         self.X_df = X_df[self.feature_params['selected']]

#         return None

#     def fix_data_split_params(self):
#         if 'random_state' not in self.data_split_params:
#             self.data_split_params['random_state'] = None
#         return None
    
#     def fix_classifier_params(self):
#         if self.classifier_params is None:
#             return None
#         if 'name' not in self.classifier_params:
#             self.classifier_params['name'] = 'XGBClassifier'
#         if 'kwargs' not in self.classifier_params:
#             self.classifier_params['kwargs'] = {}
#         if self.classifier_params['name'] in ['XGBClassifier']:
#             if 'random_state' not in self.classifier_params['kwargs']:
#                 self.classifier_params['kwargs']['random_state'] = 0
#         return None
    
#     def fix_all_params(self):
#         self.fix_feature_params()
#         self.fix_data_split_params()

#         if self.hyperopt_params is None and self.classifier_params is None:
#             self.classifier_params = {}
#         elif self.hyperopt_params is not None :
#             self.classifier_params = None

#         self.fix_classifier_params()
#         self.main_params['exp_name'] = self.get_exp_name()
#         return None
    

#     def get_exp_name(self):
        
#         exp_name = ''
        
#         prefix = self.main_params['prefix']
#         exp_name += f'[{prefix}]'
        
#         if self.data_split_params['random_state'] is not None:
#             rs = self.data_split_params['random_state']
#             exp_name += f'[ds{rs}]'

#         if clf.feature_params['to_str_func'] == 'sha256_8':
#             features_to_str_func = lambda features: get_hash(features, n = 8)
#         else:
#             features_to_str_func = self.feature_params['to_str_func']
#         feature_selected = self.feature_params['selected']
#         exp_name += f'[X_{features_to_str_func(feature_selected)}]'

#         if self.hyperopt_params is not None and len(self.hyperopt_params) > 0:
#             exp_name += f'[hyperopt_{get_hash(self.hyperopt_params, n = 8)}]'

#         if self.classifier_params is not None and len(self.classifier_params) > 0:
#             classifier_name = self.classifier_params['name']
#             classifier_kwargs = self.classifier_params['kwargs']
#             exp_name += f'[{classifier_name}_{get_hash(classifier_kwargs, n = 8)}]'

#         return exp_name

#     def run(self):
#         self.fix_all_params()
#         prefix = self.main_params['prefix']
#         results_path = self.main_params['results_path']
#         exp_name = self.get_exp_name()
#         save_name = f'{exp_name}[{get_timestamp()}]'
#         save_path = os.path.join(results_path, prefix, save_name)


#     def train(self):
#         pass

#     def predict(self):
#         pass

#     def evaluate(self):
#         pass

#     def summary(self):
#         pass

#     def save(self):
#         pass

#     pass


# X_df = pd.DataFrame({
#     'A': [1,2],
#     'B': [3,4]
# })
# y_df = X_df.copy()


# clf = BinaryClassifier(
#     X_df=X_df, y_df=y_df, 
#     main_params={
#         'prefix': 'p17-S01-sudo-fake-foobar'
#     },
#     feature_params={
#         # 'to_str_func': lambda features: '+'.join(features)
#         'to_str_func': 'sha256_8'
#     },
#     data_split_params={
#         'random_state': None,
#     },
#     hyperopt_params=None,
#     classifier_params={}
# )

# clf.fix_all_params()
# clf.get_save_name()


#%%

# train_hash = '[p16_model3][xgboost_1234abcd][]'

# params sample

# params = {
#     'data_split': {
#         'random_state': 0,
#     },
#     'feature':{
#         'selected': [],
#     },
#     'classifier':{
#         'name': 'XGBClassifier',
#         'kwargs': {
#             'max_depth': 4,
#             'learning_rate': 0.1,
#         },
#     }
# }

# params = {
#     'data_split': {
#         'random_state': 0,
#     },
#     'feature':{
#         'selected': [],
#     },
#     'hyperopt':{
#         'k_fold': {
#             'n_splits': 5,
#         },
#         'params':{},
#         'evaluator':{
#             'metric': 'roc_auc',
#             'alpha': .1,
#         }
#     }
# }

# #%%
# import json

# params = {
#     'main.prefix': 'p17-S03-dim1700',
#     'main.results_path': './path/to/resutls/',
    
#     'data_split.random_state': 0,
#     'feature.to_str_func': 'sha256_8',
#     'feature.selected': None,
#     'classifier.name': 'XGBClassifier',
#     'classifier.kwargs.max_depth': 4,
#     'classifier.kwargs.learning_rate': 0.1,
#     'classifier.kwargs.random_state': 0,
# }

# # params = {
# #     'main.prefix': 'p17-S03-dim1700',
# #     'main.results_path': './path/to/resutls/',
    
# #     'data_split.random_state': 0,
# #     'feature.to_str_func': 'sha256sum_8',
# #     'feature.selected': [],

# #     'hyperopt.k_fold': None,
# #     'hyperopt.space': None,
# #     'hyperopt.evaluator': None,
# # }



# def flat_to_nested(flat_dict: dict):
#     s = flat_dict.copy()
#     nested_dict = {}
#     for key, value in s.items():
#         parts = key.split('.')
#         current = nested_dict
#         for part in parts[:-1]:
#             if part not in current:
#                 current[part] = {}
#             current = current[part]
#         current[parts[-1]] = value
#     return nested_dict

# # 示例

# nested_dict = flat_to_nested(params)
# print(json.dumps(nested_dict, indent=4, ensure_ascii=False))

# #%%
# f = lambda : None
# def g():
#     pass

# str({
#     'f': f,
#     'g': g
#     })


# # class BinaryClassifierGridControler(object):

# #     def __init__(self,
# #                  X_df: pd.DataFrame,
# #                  y_df: pd.DataFrame,
# #                  prefix: str = 'no_prefix',
# #                  keywords: str = None,

# #                 #  features: list = None, 
# #                 features_grid: list = None,

# #                  hyperopt_params_grid: list = None,
# #                  classifier_params_grid: list = None,

                 

                 
# #                  ):
# #         pass






# #%%

# '''

# [prefix][xgboost_12345678][X_12345678][y_12345678][20140101Mon121212]

# prefix-xgboost_12345678-X_12345678-y_12345678-20140101Mon121212

# prefix-XGB_12345678-X_12345678-y_12345678-20140101Mon121212

# prefix-XGB_12345678-X_S01+S02+S03+S04_bp-y_12345678-20140101Mon121212



# '''
# #%%







